use std::collections::BTreeMap;

use anyhow::{Error, bail};
use cozo::{DbInstance, DataValue, JsonData, ScriptMutability};
use serde_json::Value;
use tracing::error;

use crate::{state::{ElementContent, ElementId, Element, types::Tag}, run_query};

pub fn set_element(db: &DbInstance, id: &ElementId, content: &ElementContent) -> anyhow::Result<()> {
    let params = vec![
        ("id", DataValue::Str(serde_json::to_string(&id)?.into())),
        ("content", DataValue::Str(serde_json::to_string(content)?.into()))
    ];

    match run_query!(&db, ":put elements {id => content}", params, ScriptMutability::Mutable) {
        Ok(_) => Ok(()),
        Err(report) => bail!(report),
    }
}

pub fn set_element_content(db: &DbInstance, id: &ElementId, content: &ElementContent) -> anyhow::Result<()> {
    let params = vec![
        ("id", DataValue::Str(serde_json::to_string(&id)?.into())),
        ("content", DataValue::Str(serde_json::to_string(content)?.into()))
    ];

    match run_query!(&db, ":put elements {id => content}", params, ScriptMutability::Mutable) {
        Ok(_) => Ok(()),
        Err(report) => bail!(report),
    }
}

pub fn remove_element(db: &DbInstance, id: &ElementId) -> anyhow::Result<()> {
    match run_query!(&db, ":delete elements {id}", vec![("id", DataValue::Str(serde_json::to_string(&id)?.into()))], cozo::ScriptMutability::Mutable) {
        Ok(_) => Ok(()),
        Err(report) => bail!(report),
    }
}

pub fn get_element(db: &DbInstance, id: &ElementId) -> anyhow::Result<Element> {
    let mut params = BTreeMap::new();
    params.insert("id".to_string(), DataValue::Str(serde_json::to_string(&id)?.into()));

    let result = db.run_script("
    ?[content] := *elements[$id, content]
    ", params, cozo::ScriptMutability::Immutable);
    match result {
        Ok(val) => {
            if let Some(firstrow) = val.rows.first() {
                if let [ DataValue::Json(JsonData(Value::String(content))) ] = firstrow.as_slice() {
                    return Ok(Element::new(
                        id.to_owned(),
                        content.as_str().try_into()?,
                    ));
                }
            }
            return Err(Error::msg("Could not parse db result as Element"));
        },
        Err(report) => bail!(report),
    }
}

pub fn get_elements_by_tag(db: &DbInstance, tag: &Tag) -> anyhow::Result<Vec<ElementId>> {
    let mut params = BTreeMap::new();
    params.insert("tag".to_string(), DataValue::Str(serde_json::to_string(tag)?.into()));

    let result = db.run_script("
    ?[element] := *tags[$tag, element]
    ", params, cozo::ScriptMutability::Immutable);

    match result {
        Ok(named_rows) => {
            let mut element_ids = vec![];
            for row in named_rows.rows {
                if let [ DataValue::Json(JsonData(Value::String(element_id)))] = row.as_slice() {
                    match serde_json::from_str(&element_id) {
                        Ok(id) => element_ids.push(id),
                        Err(e) => {
                            error!("Error parsing element id {}: {}", element_id, e);
                            continue
                        }
                    }
                }
            }
            Ok(element_ids)
        },
        Err(report) => bail!(report),
    }
}